/** An effect for non-animated meshes with
	normal-mapped diffuse and specular lighting + environment mapping. 
	Supports one light. 
	@author Jukka Jylnki */

// transformations
shared float4x4 mWorldViewProj : 	WORLDVIEWPROJ;
shared float4x4 mWorld : 	WORLD;
shared float4x4 mWorldView : 	WORLDVIEW;
shared float4x4 mView : 	VIEW;

shared float4 cDiffuse : COLOR0
<
	string UIName = "Diffuse";
> = float4(0.8f, 0.8f, 0.8f, 1.f);

shared float4 cSpecular : COLOR1
<
	string UIName = "Specular";
> = float4(0.8f, 0.8f, 0.8f, 1.f);

// Camera and light
shared float3	vCamPos			:	CAMERAP
<
	string UIName = "CameraPos";
> = float3(0,0,-10);

shared float3	vLightPos		:	LIGHTP0
<
	string UIName = "LightPos";
> = float3(0,0,-10);

shared texture tBase
<
	string UIName = "Base Texture";
>;

texture tNormalMap
<
	string UIName = "Normal Map";
>;

texture tEnvMap
<
	string UIName = "Environment Map";
>;

sampler DiffuseSampler = 
sampler_state
{
    Texture = (tBase);
    MipFilter = LINEAR;
    MinFilter = LINEAR;
    MagFilter = LINEAR;
};

sampler NormalMapSampler = 
sampler_state
{
    Texture = (tNormalMap);
    MipFilter = LINEAR;
    MinFilter = LINEAR;
    MagFilter = LINEAR;
};

sampler EnvMapSampler = 
sampler_state
{
    Texture = (tEnvMap);
    MipFilter = LINEAR;
    MinFilter = LINEAR;
    MagFilter = LINEAR;
};

struct VS_INPUT
{
	float3 pos : POSITION0;
	float2 tex : TEXCOORD0;

	float3 normal : NORMAL0;
	float3 tangent : TANGENT;
	float3 binormal : BINORMAL0;
};

struct VS_OUTPUT
{
	float4 pos : POSITION0;
	float2 tex : TEXCOORD0;

	float3 lightDir : TEXCOORD1;
	float3 viewDir : TEXCOORD2;
};

VS_OUTPUT VSMain(VS_INPUT In)
{
	VS_OUTPUT Out = (VS_OUTPUT)0;
	
	// transform pos to proj space
	Out.pos = mul(float4(In.pos,1),mWorldViewProj);
	
	// pos in world space
	float4 pos = mul(float4(In.pos,1),mWorld);
	
	// Calculate light and camera direction vectors (world space)
	float3 lightDir = vLightPos - pos.xyz;
	float3 camDir = vCamPos - pos.xyz;

	// Calculate halfway vector (between camera and light)
	float3 halfWay = camDir + lightDir;

	// convert the t/b/n from model to world space
	float3 tangent = normalize(mul(float4(In.tangent,0), mWorld)).xyz;
	float3 normal = normalize(mul(float4(In.normal,0), mWorld)).xyz;
	float3 binormal = normalize(cross(tangent, normal));
	
	// Create world-to-tangent space transformation matrix
	float4x4 worldToTangent;
	worldToTangent._11_21_31 = tangent;
	worldToTangent._12_22_32 = binormal;
	worldToTangent._13_23_33 = normal;
	worldToTangent._14_24_34 = 0;
	worldToTangent._41_42_43 = 0;
	worldToTangent._44 = 1;
	
	// Transform lightdir and halfway to tangent space
	Out.lightDir = normalize(mul(float4(lightDir,0), worldToTangent).xyz);
	Out.viewDir = normalize(mul(float4(camDir,0), worldToTangent).xyz);
	Out.tex = In.tex;

	return Out;	
}	

float4 PSMain(VS_OUTPUT In) : COLOR
{
	const float bumpSmoothness = 4.2f;
	float3 lightDir = normalize(In.lightDir);
	float3 viewDir = normalize(In.viewDir);
//	float3 normal = normalize(float3(0,0,bumpSmoothness) + 2.f * (tex2D(NormalMapSampler, In.tex) - 0.5f));
	float3 normal = normalize(tex2D(NormalMapSampler, In.tex) - 0.5f);
	normal = normalize(float3(0,0,bumpSmoothness) + normal);
	float NdotL = dot(normal, lightDir);
	float NdotV = dot(normal, viewDir);
	
	float3 reflection = normalize(2.f * normal * NdotL - lightDir); // light refl.
	float3 viewRefl = normalize(2.f * normal * NdotV - viewDir);
	float RdotV = dot(reflection, viewDir);

	if (NdotL <= 0.f)
		RdotV = 0.f;
	RdotV = max(0.f, RdotV);
	RdotV = pow(RdotV, 6.f);

	float3 diffuseColor = tex2D(DiffuseSampler, In.tex).xyz;
	float3 reflectedColor = texCUBE(EnvMapSampler, viewRefl).xyz;
	const float3 specularColor = float3(0.7f,0.7f,1);

	float diffuse = max(0.3f,  NdotL);

	return float4(reflectedColor * 0.3f + (reflectedColor + diffuseColor) * diffuse + specularColor * RdotV, 1);
}

float4 PSWhite(VS_OUTPUT In) : COLOR
{
	return float4(0.1f,0.1f,0.1f,1.f);
}

technique DefaultTechnique
{
    pass P0
    {
        VertexShader = compile vs_2_0 VSMain();
        PixelShader  = compile ps_2_0 PSMain();
    }  
}

technique EdgedFaces
{
    pass P0
    {
        VertexShader = compile vs_2_0 VSMain();
        PixelShader  = compile ps_2_0 PSMain();
    }  
    pass P1
    {
        VertexShader = compile vs_2_0 VSMain();
        PixelShader  = compile ps_2_0 PSWhite();
        DepthBias=-0.0001f;
        AlphaBlendEnable = True;
        SrcBlend = One;
        DestBlend = One;
        FillMode = WireFrame;
    }  
}