/** @file
	@author Jukka Jylnki

	This work is copyrighted material and may NOT be used for any kind of commercial or 
	personal advantage and may NOT be copied or redistributed without prior consent
	of the author(s). 
*/
static const float cShadowMapSize = 2048.f;
static const float cTexelSize = 1.f / cShadowMapSize;
   
// Transforms the fragment pos from light's projection space to shadowmap's texture space.
float2 LightProjPosToShadowMapPos(float4 vLightProjPos)
{
	float2 ShadowTexC = 0.5f * vLightProjPos.xy / vLightProjPos.w + float2(0.5, 0.5);
	ShadowTexC.y = 1.0f - ShadowTexC.y;
	return ShadowTexC;
}

float SampleShadowMap4TapFuzzyEval(float4 vLightProjPos, sampler sShadowMap)
{
/* For Fuzzy evaluation:
	escapes:
	static const float cRampInterval = 1.f/3e-3f;
	static const float cShadowMapEpsilon = 1.83e-3f + 1.f/cRampInterval;

	slightly swims:
	static const float cRampInterval = 1.f/1e-3f;
	static const float cShadowMapEpsilon = 2e-3f + 1.f/cRampInterval;

	pretty good:
	static const float cRampInterval = 1.f/1.5e-3f;
	static const float cShadowMapEpsilon = 2e-3f + 1.f/cRampInterval; 
*/
	static const float cRampInterval = 1.f/1.3e-3f;
	static const float cShadowMapEpsilon = 2e-3f + 1.f/cRampInterval;

	// shadowmap's texel coordinate
	float2 ShadowTexC = LightProjPosToShadowMapPos(vLightProjPos);

	// Sample the lightmap four times to get how far the light reaches in the direction of this pixel.
	float4 lightReach;

	static const float2 ofsX = float2(cTexelSize, 0);
	static const float2 ofsY = float2(0, cTexelSize);

	lightReach.x = tex2D(sShadowMap, ShadowTexC - ofsX).x;
	lightReach.y = tex2D(sShadowMap, ShadowTexC + ofsX).x;
	lightReach.z = tex2D(sShadowMap, ShadowTexC - ofsY).x;
	lightReach.w = tex2D(sShadowMap, ShadowTexC + ofsY).x;

	// fragment is this far from the light source (in light's projspace)
	const float lightDist = vLightProjPos.z / vLightProjPos.w - cShadowMapEpsilon; 

	// Smooth ramp method: (Fuzzy depth comparison, Eliminating Surface Acne with Gradient Shadow Mapping, Christian Schler, ShaderX4)
	float lightAmount = 0.25f * 
		(saturate((lightReach.x - lightDist) * cRampInterval)
		+saturate((lightReach.y - lightDist) * cRampInterval)
		+saturate((lightReach.z - lightDist) * cRampInterval)
		+saturate((lightReach.w - lightDist) * cRampInterval));

	// if outside texture or behind light:
	if (ShadowTexC.x <= 0.f) lightAmount = 0.f;
	if (ShadowTexC.y <= 0.f) lightAmount = 0.f;
	if (ShadowTexC.x >= 1.f) lightAmount = 0.f;
	if (ShadowTexC.y >= 1.f) lightAmount = 0.f;
	if (vLightProjPos.w < 0.f) lightAmount = 0.f;

	return lightAmount;
}

float SampleShadowMap4TapStepEval(float4 vLightProjPos, sampler sShadowMap)
{
/*  For standard step evaluation:
	3.0e-3f seems to be really too much, the shadows escape.
	2.2e-3f is too little, aliasing occurs.
	2.25e-3f seems to be good for step method (1).
	Tested on projmatrix with fov=1.4,front=5,back=200,res=2048,aspect=1  */
	static const float cShadowMapEpsilon = 2.25e-3f;

	// shadowmap's texel coordinate
	float2 ShadowTexC = LightProjPosToShadowMapPos(vLightProjPos);

	// Sample the lightmap four times to get how far the light reaches in the direction of this pixel.
	float4 lightReach;

	static const float2 ofsX = float2(cTexelSize, 0);
	static const float2 ofsY = float2(0, cTexelSize);

	lightReach.x = tex2D(sShadowMap, ShadowTexC - ofsX).x;
	lightReach.y = tex2D(sShadowMap, ShadowTexC + ofsX).x;
	lightReach.z = tex2D(sShadowMap, ShadowTexC - ofsY).x;
	lightReach.w = tex2D(sShadowMap, ShadowTexC + ofsY).x;

	// fragment is this far from the light source (in light's projspace)
	const float lightDist = vLightProjPos.z / vLightProjPos.w - cShadowMapEpsilon; 

	// The standard step method:
	float lightAmount = 
	     ((lightReach.x < lightDist) ? 0 : 0.25f) +
	     ((lightReach.y < lightDist) ? 0 : 0.25f) +
	     ((lightReach.z < lightDist) ? 0 : 0.25f) +
	     ((lightReach.w < lightDist) ? 0 : 0.25f);

	// if outside texture or behind light:
	if (ShadowTexC.x <= 0.f) lightAmount = 0.f;
	if (ShadowTexC.y <= 0.f) lightAmount = 0.f;
	if (ShadowTexC.x >= 1.f) lightAmount = 0.f;
	if (ShadowTexC.y >= 1.f) lightAmount = 0.f;
	if (vLightProjPos.w < 0.f) lightAmount = 0.f;

	return lightAmount;
}

float SampleShadowMap1TapStepEval(float4 vLightProjPos, sampler sShadowMap)
{
/*  For standard step evaluation:
	3.0e-3f seems to be really too much, the shadows escape.
	2.2e-3f is too little, aliasing occurs.
	2.25e-3f seems to be good for step method (1).
	Tested on projmatrix with fov=1.4,front=5,back=200,res=2048,aspect=1 */
	static const float cShadowMapEpsilon = 2.25e-3f;

	// shadowmap's texel coordinate
	float2 ShadowTexC = LightProjPosToShadowMapPos(vLightProjPos);

	// Sample the lightmap four times to get how far the light reaches in the direction of this pixel.
	float lightReach = tex2D(sShadowMap, ShadowTexC).x;

	// fragment is this far from the light source (in light's projspace)
	const float lightDist = vLightProjPos.z / vLightProjPos.w - cShadowMapEpsilon; 

	// The standard step method:
	float lightAmount = step(lightDist, lightReach);

	// if outside texture or behind light:
//	if (ShadowTexC.x <= 0.f || ShadowTexC.y <= 0.f || 
//		ShadowTexC.x >= 1.f || ShadowTexC.y >= 1.f || vLightProjPos.w < 0.f)
//		lightAmount = 0.f;

	if (ShadowTexC.x <= 0.f) lightAmount = 0.f;
	if (ShadowTexC.y <= 0.f) lightAmount = 0.f;
	if (ShadowTexC.x >= 1.f) lightAmount = 0.f;
	if (ShadowTexC.y >= 1.f) lightAmount = 0.f;
	if (vLightProjPos.w < 0.f) lightAmount = 0.f;

	return lightAmount;
}