/** @file
	@author Jukka Jylnki

	This work is copyrighted material and may NOT be used for any kind of commercial or 
	personal advantage and may NOT be copied or redistributed without prior consent
	of the author(s). 
*/

static const bool g_bSpecular = true;
static const bool g_bNormalMapping = true;

string ParamID = "0x00";

#include "ShadowMap.fx"

// transformations
shared float4x4 mWorldViewProj : 	WORLDVIEWPROJ; //: WORLDVIEWPROJECTION;
shared float4x4 mWorld : 	WORLD;
shared float4x4 mWorldView : 	WORLDVIEW;
shared float4x4 mView : 	VIEW;

shared float4 cDiffuse : COLOR0 //: DIFFUSE0
<
	string UIName = "Diffuse";
> = float4(0.8f, 0.8f, 0.8f, 1.f);

shared float4 cSpecular : COLOR1 //: SPECULAR0
<
	string UIName = "Specular";
> = float4(0.8f, 0.8f, 0.8f, 1.f);

float fSpecularPower <
	string UIName = "Specular Power";
	float UIMin = 0.0f;
	float UIMax = 128.0f;
	float UIStep = 0.1f;
>   = 1.0f;

float fReflectivity <
	string UIName = "Reflectivity";
	float UIMin = 0.0f;
	float UIMax = 1.0f;
	float UIStep = 0.01f;
>   = 1.0f;

// Camera and light
shared float3	vCamPos			 : CAMERAP	:	WORLDCAMERAPOSITION;

shared float3	vLightPos	 : LIGHTP0	:	Position
<
	string UIName = "Light"; 
	string Object = "TargetLight";
	int refID = 0;
>;

float fBumpSmoothness <
	string UIName = "Bump Smoothness";
	float UIMin = 0.0f;
	float UIMax = 5.0f;
	float UIStep = 0.01f;
>   = 1.0f;

shared texture tBase
<
	string UIName = "Base Texture";
>;

texture tNormalMap
<
	string UIName = "Normal Map";
>;

texture tEnvMap
<
	string UIName = "Environment Map";
>;

shared texture tShadowMap			:	SHADOWMAP;
shared float4x4 mShadowMapViewProj	: 	SHADOWMAPVIEWPROJ;

sampler DiffuseSampler = 
sampler_state
{
    Texture = (tBase);
    MipFilter = LINEAR;
    MinFilter = LINEAR;
    MagFilter = LINEAR;
};

sampler NormalMapSampler = 
sampler_state
{
    Texture = (tNormalMap);
    MipFilter = LINEAR;
    MinFilter = LINEAR;
    MagFilter = LINEAR;
};

sampler EnvMapSampler = 
sampler_state
{
    Texture = (tEnvMap);
    MipFilter = LINEAR;
    MinFilter = LINEAR;
    MagFilter = LINEAR;
};

sampler ShadowMapSampler = 
sampler_state
{
    Texture = (tShadowMap);
    MipFilter = LINEAR;
    MinFilter = LINEAR;
    MagFilter = LINEAR;
    AddressU = Clamp;
    AddressV = Clamp;
};

struct VS_INPUT
{
	float3 pos : POSITION0;
	float2 tex : TEXCOORD0;

	float3 normal : NORMAL;
	float3 tangent : TANGENT;
	float3 binormal : BINORMAL0;
};

struct VS_OUTPUT
{
	float4 pos : POSITION0;
	float2 tex : TEXCOORD0;
	float4 vLightProjPos : TEXCOORD1; // vertex pos in light's projected space
	
	float3 lightDir : TEXCOORD2;
	float3 viewDir : TEXCOORD3;
};

VS_OUTPUT VSMain(VS_INPUT In)
{
	VS_OUTPUT Out = (VS_OUTPUT)0;
	
	// transform pos to proj space
	Out.pos = mul(float4(In.pos,1),mWorldViewProj);
	
	// pos in world space
	float4 pos = mul(float4(In.pos,1),mWorld);
	
	// Calculate light and camera direction vectors (world space)
	float3 lightDir = vLightPos - pos.xyz;
	float3 camDir = vCamPos - pos.xyz;

	// Calculate halfway vector (between camera and light)
//	float3 halfWay = camDir + lightDir;

	// convert the t/b/n from model to world space
	float3 tangent = normalize(mul(float4(In.tangent,0), mWorld)).xyz;
	float3 normal = normalize(mul(float4(In.normal,0), mWorld)).xyz;
	float3 binormal = normalize(cross(tangent, normal));
	
	// Create world-to-tangent space transformation matrix
	float4x4 worldToTangent;
	worldToTangent._11_21_31 = tangent;
	worldToTangent._12_22_32 = binormal;
	worldToTangent._13_23_33 = normal;
	worldToTangent._14_24_34 = 0;
	worldToTangent._41_42_43 = 0;
	worldToTangent._44 = 1;
	
	// Transform lightdir and halfway to tangent space
	Out.lightDir = normalize(mul(float4(lightDir,0), worldToTangent).xyz);
	Out.viewDir = normalize(mul(float4(camDir,0), worldToTangent).xyz);
//	Out.lightDir = normalize(lightDir);//mul(float4(lightDir,0), worldToTangent).xyz);
//	Out.viewDir = normalize(camDir);//mul(float4(camDir,0), worldToTangent).xyz);
	Out.tex = In.tex;

	// Transform pos to light view space.
	Out.vLightProjPos = mul(pos, mShadowMapViewProj);
	
	return Out;	
}	

//static const bool g_bSpecular = true;
//static const bool g_bNormalMapping = true;

float3 SampleNormalMap(float2 texCoord, sampler sNormalMap)
{
	return normalize(2.f * (tex2D(sNormalMap, texCoord) - 0.5f));
}

float3 SampleNormalMap(float2 texCoord, sampler sNormalMap, float bumpSmoothness)
{
	return normalize(float3(0,0,bumpSmoothness) + 
			2.f * (tex2D(sNormalMap, texCoord) - 0.5f));
}

float4 PSMain(VS_OUTPUT In) : COLOR
{
//	float lightAmount = SampleShadowMap(In.vLightProjPos, ShadowMapSampler);

	float3 normal;
	if (g_bNormalMapping)
		normal = SampleNormalMap(In.tex, NormalMapSampler, fBumpSmoothness);
	else
		normal = float3(0,0,1);

	float3 viewDir = normalize(In.viewDir);
	float3 lightDir = normalize(In.lightDir);

	// Diffuse mapping
	float3 diffuseColor = tex2D(DiffuseSampler, In.tex).xyz * cDiffuse.xyz;
	float NdotL = max(0.f,dot(normal, lightDir));

	// Environment mapping
	float NdotV = dot(normal, viewDir);

	// reflect view direction vector by the normal
	float3 viewRefl = normalize(2.f * normal * NdotV - viewDir); 
	float3 reflectedColor = texCUBE(EnvMapSampler, viewRefl).xyz;

	if (g_bSpecular)
	{
		// reflect light direction vector by the normal
		float3 reflection = normalize(2.f * normal * NdotL - lightDir);
		float RdotV = dot(reflection, viewDir);
		if (NdotL <= 0.f)// || lightAmount <= 0.5f) // no speculars if no diffuse or if in shadow
			RdotV = 0.f;
		RdotV = max(0.f, RdotV);
		RdotV = pow(RdotV, fSpecularPower);

		return float4(reflectedColor * fReflectivity + diffuseColor * NdotL + cSpecular.xyz * RdotV, 1);
	}
	else
		return float4(reflectedColor * fReflectivity + diffuseColor * NdotL, 1);
}

float4 PSWhite(VS_OUTPUT In) : COLOR
{
	return float4(0.1f,0.1f,0.1f,1.f);
}

technique DefaultTechnique
{
    pass P0
    {
        VertexShader = compile vs_2_0 VSMain();
        PixelShader  = compile ps_2_0 PSMain();
    }  
}
