/** @file
	@author Jukka Jylnki

	This work is copyrighted material and may NOT be used for any kind of commercial or 
	personal advantage and may NOT be copied or redistributed without prior consent
	of the author(s). 
*/

string ParamID = "0x00";

#include "ShadowMap.fx"

#ifdef SKINNING
float4x4 mBoneTM[cMaxBones] : BONEPALETTE;
#include "skinning.fx"
#endif

// transformations
shared float4x4 mWorldViewProj	:	WORLDVIEWPROJ;
shared float4x4 mWorld	:	WORLD;
shared float4x4 mWorldView	:	WORLDVIEW;
shared float4x4 mView	:	VIEW;

shared float4 cDiffuse : COLOR0
<
	string UIName = "Diffuse";
> = float4(0.8f, 0.8f, 0.8f, 1.f);

shared float4 cSpecular : COLOR1
<
	string UIName = "Specular";
> = float4(0.8f, 0.8f, 0.8f, 1.f);

float fSpecularPower <
	string UIName = "Specular Power";
	float UIMin = 0.0f;
	float UIMax = 128.0f;
	float UIStep = 0.1f;
>   = 1.0f;

float fReflectivity <
	string UIName = "Reflectivity";
	float UIMin = 0.0f;
	float UIMax = 1.0f;
	float UIStep = 0.01f;
>   = 1.0f;

// Camera and light
shared float3	vCamPos				:	CAMERAP;

shared float3	vLightPos			:	LIGHTP0
<
	string UIName = "Light"; 
	string Object = "TargetLight";
	int refID = 0;
>;

#ifdef SPOT_LIGHT
shared float3	vLightFacing			:	LIGHTF0;
#endif

shared float vLightAttenuation : LIGHTATTENUATION;

float fBumpSmoothness <
	string UIName = "Bump Smoothness";
	float UIMin = 0.0f;
	float UIMax = 5.0f;
	float UIStep = 0.01f;
>   = 1.0f;

shared texture tBase
<
	string UIName = "Base Texture";
>;

texture tNormalMap
<
	string UIName = "Normal Map";
>;

texture tEnvMap
<
	string UIName = "Environment Map";
>;

sampler DiffuseSampler = 
sampler_state
{
    Texture = (tBase);
    MipFilter = LINEAR;
    MinFilter = LINEAR;
    MagFilter = LINEAR;
};

sampler NormalMapSampler = 
sampler_state
{
    Texture = (tNormalMap);
    MipFilter = LINEAR;
    MinFilter = LINEAR;
    MagFilter = LINEAR;
};

sampler EnvMapSampler = 
sampler_state
{
    Texture = (tEnvMap);
    MipFilter = LINEAR;
    MinFilter = LINEAR;
    MagFilter = LINEAR;
};

#ifdef SHADOW_MAPPING
shared texture tShadowMap	:	SHADOWMAP;
shared float4x4 mShadowMapViewProj	:	SHADOWMAPVIEWPROJ;

sampler ShadowMapSampler = 
sampler_state
{
    Texture = (tShadowMap);
    MipFilter = Point;
    MinFilter = Point;
    MagFilter = Point;
    AddressU = Clamp;
    AddressV = Clamp;
};
#endif

struct VS_INPUT
{
	float3 pos : POSITION0;
	float2 tex : TEXCOORD0;

#ifdef SKINNING
	float3 BlendWeights    : BLENDWEIGHT;
	float4 BlendIndices    : BLENDINDICES;
#endif

	float3 normal : NORMAL;
	float3 tangent : TANGENT;
	float3 binormal : BINORMAL0;
};

struct VS_OUTPUT
{
	float4 pos : POSITION0;
	float2 tex : TEXCOORD0;
	float4 vLightProjPos : TEXCOORD1; // vertex pos in light's projected space
	
	float3 lightDir : TEXCOORD2;
	float3 viewDir : TEXCOORD3;
	float lightDist : TEXCOORD4;
#ifdef SPOT_LIGHT
	float3 lightFacing : TEXCOORD5;
#endif
};

VS_OUTPUT VSMain(VS_INPUT In)
{
	VS_OUTPUT Out = (VS_OUTPUT)0;
	
	// transform pos to proj space
#ifdef SKINNING
	float4 p = SkinVertex(In.BlendWeights, In.BlendIndices, float4(In.pos.xyz, 1));
	Out.pos = mul(p, mWorldViewProj);
	// pos in world space
	float4 pos = mul(p, mWorld);
#else
	Out.pos = mul(float4(In.pos,1), mWorldViewProj);
	// pos in world space
	float4 pos = mul(float4(In.pos,1), mWorld);
#endif
		
	// Calculate light and camera direction vectors (world space)
	float3 lightDir = vLightPos - pos.xyz;
	float3 camDir = vCamPos - pos.xyz;

	// convert the t/b/n from model to world space
	float3 tangent = normalize(mul(float4(In.tangent,0), mWorld)).xyz;
	float3 normal = normalize(mul(float4(In.normal,0), mWorld)).xyz;
	float3 binormal = normalize(cross(tangent, normal));
	
	// Create world-to-tangent space transformation matrix
	float4x4 worldToTangent;
	worldToTangent._11_21_31 = tangent;
	worldToTangent._12_22_32 = binormal;
	worldToTangent._13_23_33 = normal;
	worldToTangent._14_24_34 = 0;
	worldToTangent._41_42_43 = 0;
	worldToTangent._44 = 1;
	
	// Transform lightdir and viewdir to tangent space
	Out.lightDir = normalize(mul(float4(lightDir,0), worldToTangent).xyz);
	Out.viewDir = normalize(mul(float4(camDir,0), worldToTangent).xyz);
	Out.lightDist = length(vLightPos - pos.xyz);

#ifdef SPOT_LIGHT
	Out.lightFacing = normalize(mul(float4(vLightFacing,0), worldToTangent).xyz);
#endif

	// Just pass UVs without modification
	Out.tex = In.tex;

#ifdef SHADOW_MAPPING
	// Transform pos to light view space.
	Out.vLightProjPos = mul(pos, mShadowMapViewProj);
#endif

	return Out;	
}	

/** Samples a texel from a normal map and converts the value into a proper direction vector.
	@return Normal vector stored in the given position at the normal map. */
float3 SampleNormalMap(float2 texCoord, sampler sNormalMap)
{
	return normalize(2.f * (tex2D(sNormalMap, texCoord) - 0.5f));
}

/** Samples a texel from a normal map and converts the value into a proper direction vector. Finally,
	a custom smoothing factor is applied.
	@param bumpSmoothness The amount of smoothing to apply to the directional vector. Larger == more.
	@return Normal vector stored in the given position at the normal map. */
float3 SampleNormalMap(float2 texCoord, sampler sNormalMap, float bumpSmoothness)
{
	return normalize(float3(0,0,bumpSmoothness) + 
			2.f * (tex2D(sNormalMap, texCoord) - 0.5f));
}

float4 PSMain(VS_OUTPUT In) : COLOR
{
#ifdef SHADOW_MAPPING
	#ifdef SHADOW_MAPPING_AMBIENT
		#ifdef SM20
			float lightAmount = saturate(g_ambientLight + SampleShadowMap1TapStepEval(In.vLightProjPos, ShadowMapSampler));
		#else
			float lightAmount = saturate(g_ambientLight + SampleShadowMap4TapFuzzyEval(In.vLightProjPos, ShadowMapSampler));
		#endif
	#else
		#ifdef SM20
			float lightAmount = SampleShadowMap1TapStepEval(In.vLightProjPos, ShadowMapSampler);
		#else
			float lightAmount = SampleShadowMap4TapFuzzyEval(In.vLightProjPos, ShadowMapSampler);
		#endif
	#endif
#endif

	float3 normal;
	if (g_bNormalMapping)
#ifdef SM20
		normal = SampleNormalMap(In.tex, NormalMapSampler);
#else
		normal = SampleNormalMap(In.tex, NormalMapSampler, fBumpSmoothness);
#endif
	else
		normal = float3(0,0,1);

	float3 viewDir = normalize(In.viewDir);
	float3 lightDir = normalize(In.lightDir);

	// Diffuse mapping
	float3 diffuseColor = tex2D(DiffuseSampler, In.tex).xyz * cDiffuse.xyz;
#ifdef SPOT_LIGHT
	float3 lightFacing = normalize(In.lightFacing);
	float umbra = saturate(dot(lightDir, lightFacing) * 5.f - 3.5f);
	umbra = pow(umbra, 4.f);
	float NdotL = max(0.f,dot(normal, lightDir) * umbra * min(2.f, vLightAttenuation / In.lightDist));
#else
	float NdotL = max(0.f,dot(normal, lightDir) * vLightAttenuation / In.lightDist);
#endif

	float3 viewRefl;
	float3 reflectedColor;
	if (g_bEnvMapping)
	{
		// Environment mapping
		float NdotV = dot(normal, viewDir);

		// reflect view direction vector by the normal
		viewRefl = normalize(2.f * normal * NdotV - viewDir); 
		reflectedColor = texCUBE(EnvMapSampler, viewRefl).xyz;
	}
	
	if (g_bSpecular)
	{
		// reflect light direction vector by the normal
		float3 reflection = normalize(2.f * normal * NdotL - lightDir);
		float RdotV = dot(reflection, viewDir);
#ifdef SHADOW_MAPPING
		RdotV = min(RdotV, lightAmount);
//		if (lightAmount <= 0.f) // no speculars if no diffuse or if in shadow
//			RdotV = 0.f;
#endif
		RdotV = min(RdotV, NdotL);
//		if (NdotL <= 0.f) // no speculars if no diffuse
//			RdotV = 0.f;

		RdotV = max(0.f, RdotV);
		RdotV = pow(RdotV, fSpecularPower);
#ifdef SHADOW_MAPPING
		if (g_bEnvMapping)
			return float4(lightAmount * (NdotL * (reflectedColor * fReflectivity + diffuseColor) + cSpecular.xyz * RdotV), 1);
		else
			return float4(lightAmount * (diffuseColor * NdotL + cSpecular.xyz * RdotV), 1);
#else
		if (g_bEnvMapping)
			return float4(NdotL * (reflectedColor * fReflectivity + diffuseColor) + cSpecular.xyz * RdotV, 1);
		else
			return float4(diffuseColor * NdotL + cSpecular.xyz * RdotV, 1);
#endif
	}
	else
#ifdef SHADOW_MAPPING
		if (g_bEnvMapping)
			return float4(lightAmount * NdotL * (reflectedColor * fReflectivity + diffuseColor), 1);
		else
			return float4(lightAmount * diffuseColor * NdotL, 1);
#else
		if (g_bEnvMapping)
			return float4(reflectedColor * fReflectivity * NdotL + diffuseColor * NdotL, 1);
		else
			return float4(diffuseColor * NdotL, 1);
#endif
}

technique DefaultTechnique
{
    pass P0
    {
#ifdef SM20
        VertexShader = compile vs_2_0 VSMain();
        PixelShader  = compile ps_2_0 PSMain();
#else
        VertexShader = compile vs_3_0 VSMain();
        PixelShader  = compile ps_3_0 PSMain();
#endif
    }  
}
